import numpy as np
from utilities import Utilities, equation_type_converter
import pandas as pd
from simplex import Simplex
from aliases import Variables, Iterations
from math import pow


class BigM(Simplex):

    def __init__(self) -> None:
        self.m = pow(10, 18)

    def __clean_preconditioned_df(self, preconditioned_df: pd.DataFrame, rows: list[int], columns: list[str]) -> pd.DataFrame:
        # Dropping rows
        preconditioned_df.drop(rows, inplace=True)

        # Droping columns
        preconditioned_df.drop(columns, axis=1, inplace=True)

        return preconditioned_df

    def __handle_negative_constraints(self, preconditioned_matrix: np.matrix) -> np.matrix:
        # print("before modification")
        # print(preconditioned_matrix)
        # input()
        for j in range(0, preconditioned_matrix.shape[1] - 1):
            if preconditioned_matrix.item((preconditioned_matrix.shape[0] - 2, j)) < 0:
                for i in range(0, preconditioned_matrix.shape[0] - 1):
                    preconditioned_matrix.itemset(
                        (i, j), preconditioned_matrix.item((i, j)) * -1)
                preconditioned_matrix.itemset(
                    (preconditioned_matrix.shape[0] - 1, j), equation_type_converter[preconditioned_matrix.item((preconditioned_matrix.shape[0] - 1, j))])

        # print("after modification")
        # print(preconditioned_matrix)
        # input()

        return preconditioned_matrix

    def __determine_coefs(self, preconditioned_matrix: np.matrix) -> list[str]:
        coefs = []
        x = 0
        flag = True

        for i in range(0, preconditioned_matrix.shape[0]):
            if (preconditioned_matrix.item(i, preconditioned_matrix.shape[1] - 1)) != -1 and flag:
                flag = False

            if (preconditioned_matrix.item(i, preconditioned_matrix.shape[1] - 1)) == -1 and flag:
                x += 1
                coefs.append(f'x{x}')
            elif (preconditioned_matrix.item(i, preconditioned_matrix.shape[1] - 1)) == 0:
                si = Utilities.get_col_index(preconditioned_matrix, i)
                coefs.append(f's{si +1}')
            elif (preconditioned_matrix.item(i, preconditioned_matrix.shape[1] - 1)) == 1:
                ai = Utilities.get_col_index(preconditioned_matrix, i)
                coefs.append(f'a{ai +1}')
            elif (preconditioned_matrix.item(i, preconditioned_matrix.shape[1] - 1)) == 2:
                coefs.append("p")

        return coefs

    def __transform_equation(self, preconditioned_matrix: np.matrix) -> np.matrix:
        # Multiply all the coefs of the equations by -1
        for i in range(0, preconditioned_matrix.shape[0]):
            preconditioned_matrix.itemset(
                (i, preconditioned_matrix.shape[1]-2), preconditioned_matrix.item((i, preconditioned_matrix.shape[1]-2)) * -1)

        # Add the P coef
        p_row = [0] * (preconditioned_matrix.shape[1])

        p_row[len(p_row) - 2] = 1
        p_row[len(p_row) - 1] = 2  # Flag hat stands a p coef
        preconditioned_matrix = np.insert(
            preconditioned_matrix, preconditioned_matrix.shape[0] - 2, p_row, axis=0)

        return preconditioned_matrix

    def __preconditioner(self, formattedInput: np.matrix) -> pd.DataFrame:
        # Adding the artificial and slack variables to the formattedInput matrix
        artificial_variables = []
        slack_variables = []
        for j in range(0, formattedInput.shape[1] - 1):
            item = formattedInput.item((formattedInput.shape[0] - 1, j))
            if item >= 0:
                artificial_variables.append(1)
            else:
                artificial_variables.append(0)

            if item == 0:
                slack_variables.append(0)
            elif item > 0:
                slack_variables.append(-1)
            else:
                slack_variables.append(+1)

        # print("Slack variables:" + str(slack_variables))
        # print("Artificial variables:" + str(artificial_variables))

        # Copy the formattedInput to the preconditioned_matrix (data is backuped in the formattedInput + avoid refrence access)
        preconditioned_matrix = formattedInput.copy()

        # Handle negative constraints
        preconditioned_matrix = self.__handle_negative_constraints(preconditioned_matrix)

        # Initialize a new column with -1 values (by default) to determine later on
        # if a variable is a slack or artificial variable
        variable_flag_col = []
        for i in range(0, formattedInput.shape[0]):
            variable_flag_col.append([-1])
        preconditioned_matrix = np.append(
            preconditioned_matrix, variable_flag_col, axis=1)

        # print("1st version of the preconditioned matrix:")
        # print(preconditioned_matrix)
        # input()

        # Initilize default row that is going to modified based on the
        # the slack and artificial variables positions in the system
        new_row = []
        for j in range(0, preconditioned_matrix.shape[1]):
            new_row.append(0)

        # Handling slack variables
        for k in range(0, len(slack_variables)):
            slack_variable_existence = False
            if slack_variables[k] != 0:
                slack_variable_existence = True
                new_row[k] = slack_variables[k]

            if k != 0:
                new_row[k - 1] = 0

            if slack_variable_existence:
                preconditioned_matrix = np.insert(
                    preconditioned_matrix, preconditioned_matrix.shape[0] - 2, new_row, axis=0)
                preconditioned_matrix.itemset(
                    (preconditioned_matrix.shape[0] - 1, k), 0)
        new_row[k] = 0

        # Handling artificial variables
        new_row[len(new_row) - 1] = 1
        for k in range(0, len(artificial_variables)):
            artificial_variable_existence = False
            if artificial_variables[k] != 0:
                artificial_variable_existence = True
                new_row[k] = artificial_variables[k]

            if k != 0:
                new_row[k - 1] = 0

            if artificial_variable_existence:
                preconditioned_matrix = np.insert(
                    preconditioned_matrix, preconditioned_matrix.shape[0] - 2, new_row, axis=0)

        # print("2sec version of the preconditioned matrix:")
        # print(preconditioned_matrix)
        # input()

        # Adding the P coef + making modifications to max / min equation
        preconditioned_matrix = self.__transform_equation(
            preconditioned_matrix)

        # print("Equation modifications")
        # print(preconditioned_matrix)

        # Determining the coefs within our system after adding
        # the slack and artificial variables
        coefs = self.__determine_coefs(preconditioned_matrix)
        # print(coefs)

        # Converting to pandas df for ease of manipulation
        # and pretty display in jupiter notebook
        preconditioned_matrix_t = preconditioned_matrix.transpose()

        # print("The transposed matrix")
        # print(preconditioned_matrix_t)

        preconditioned_df = pd.DataFrame(preconditioned_matrix_t, columns=[
                                         *coefs, "condition", "flags"])

        # print("The dataframe")
        # print(preconditioned_df)

        filtred_cols = preconditioned_df.filter(regex="a\d").columns

        # print("Filtered cols")
        # print(filtred_cols)

        for col in filtred_cols:
            preconditioned_df.loc[preconditioned_df.shape[0] -
                                  2, [col]] = self.m

        # Drop columns and delete rows that are unnecessary
        preconditioned_df = self.__clean_preconditioned_df(
            preconditioned_df, [preconditioned_df.shape[0] - 1], ["flags"])

        # print("The dataframe after cleaning")
        # print(preconditioned_df)

        return preconditioned_df

    # This method will perform operations between rows to get
    # rid of the Ms in the artificial variables columns

    def __prepare_matrix(self, preconditioned_df: pd.DataFrame) -> pd.DataFrame:
        """
        This method performs operations between rows of the preconditioned DataFrame
        to get rid of every M that resides in one of the artificial variables columns.
        Arguments:
            preconditioned_df: A pandas DataFrame that holds the linear program
            after adding slack and artificial variables, performing transformations on the 
            eqution of maximisation and cleaning the DataFrame from unnecessary data
        Returns:
            A pandas DataFrame that allows the execution of the simplex algorithm
        """
        objectiveFunctionRow = preconditioned_df.shape[0] - 1
        newChange = True
        while newChange:  # idk if that's necessary but meh
            newChange = False
            # loop through the artificial variables columns
            for col in preconditioned_df.filter(regex="a\d").columns:
                # if last rows column of artificial variable contains M
                if preconditioned_df.loc[objectiveFunctionRow][col] != 0:
                    for row in range(0, preconditioned_df.shape[0] - 1):
                        if preconditioned_df.loc[row][col] != 0:
                            # we can perform operations between this row and the last row
                            # to get rid of the M
                            preconditioned_df = self.__perform_operations(
                                preconditioned_df, row, objectiveFunctionRow, col)
                            newChange = True

        return preconditioned_df

    def __perform_operations(self, preconditioned_df: pd.DataFrame, row: int, objectiveFunctionRow: int, col: str) -> pd.DataFrame:
        """
        This method performs operations between row and the objective function row
        to get rid of the M in the artificial variable column that resides in the column col.
        Arguments:
            preconditioned_df: A pandas DataFrame that holds the linear program.
            row: The row that we can use to perform operations with the objective function row.
            objectiveFunctionRow: The row that contains the objective function
            col: The column that contains the M
        Returns:
            A pandas DataFrame after performing the operations.
        """
        operation = -preconditioned_df.loc[objectiveFunctionRow][col] / \
            preconditioned_df.loc[row][col]  # quick meth, trivial
        # loop through the columns to do chaka laka boom boom between the rows
        for column in preconditioned_df.columns:
            preconditioned_df.loc[objectiveFunctionRow][column] += operation * \
                preconditioned_df.loc[row][column]
        return preconditioned_df

    def runBigM(self, formattedInput: np.matrix) -> Iterations:
        """
        This function solves linear optimisation problems with the Big M method and returns each iteration
        in a seperate matrix.
        Arguments:
            formattedInput: A description matrix generated by BigM.formatUserInput().
        Returns:
            A list of matrices that corresponds to simplex iterations (+ the Big M initial phase of course)
        """
        # try:
        # Formatted input preconditioning
        # print("formated input:")
        # print(formattedInput)
        # print()
        # input('')

        preconditioned_df = self.__preconditioner(formattedInput)
        # print("Preconditioned DataFrame")
        # print(preconditioned_df)
        # print()
        # input('')

        # Preparing the matrix for the simplex algorithm
        init_simplex_df = self.__prepare_matrix(preconditioned_df)
        # print("initial simplex matrix")
        # print(init_simplex_df)
        # print()
        # input()

        # Running the normal simplex algorithm on the matrix
        iterations = super()._perform_simplex(init_simplex_df)

        return iterations
        # except:
        #     print("oops")
        #     return []



import numpy as np


class Utilities:

    @staticmethod
    def get_col_index(matrix: np.matrix, i: int) -> int:
        ix = -1
        for j in range(0, matrix.shape[1]):
            if matrix.item(i, j) != 0:
                ix = j
                break
        return ix

    @staticmethod
    def handle_type_detection(x: any):
        if type(x) == str:
            x = x[1: len(x)]
            return float(x) * -1
        else:
            return x

    @staticmethod
    def debug(entity: any, msg: str = None):
        print()
        if msg != None:
            print(msg)
        print(entity)
        print("*************************************")
        input('press any key to continue.')
        print()


equation_type_converter = {
    1: -2,
    2: -1,
    -1: 2,
    -2: 1,
    0: 0
}



#Cette fonction s'execute pour analyser les donner qui provient de chaque itération du big M
#bien sur à l'excution de la fontion les valeurs de retours ne vont pas s'afficher car il faut avoir des input correcte (qui sont en commentaires si vous voulez testez)

#RQ : cette fonction recoie en input une iteration ( de la fonction de Med Amine) qui contient les variables de bases , hors bases et une matrice du big M
# Pour les variables de bases et hors bases , ils seront tout simplement afficher et pour la matrice de l'iteration on affiche le z max


def AnalyseData(variablesdebases , variableshorsbases , mat):
    """
    Not Tested
    """
    print("les variables de bases dans cette iteration sont")
    for var in enumerate(variablesdebases):
        nomvariable=var[1].get('name')
        valeurvariable=var[1].get('value')
        print(nomvariable+'='+str(valeurvariable))
    print("les variables hors bases dans cette iteration sont")
    for var in enumerate(variableshorsbases):
        nomvariable=var[1].get('name')
        valeurvariable=var[1].get('value')
        print(nomvariable+'='+str(valeurvariable))
    print(-mat[[n-1],[m-1]])
    return 0

# Une variable correspond au couple (nom , valeur) , exemple: {"name": "x1": "value": 3.0}
#declaration de "Variable"
Variable = dict[str, float]                         

#des exmples de Variables(juste pour le test)
#Variable1={"name": "x", "value": 3.0}
#Variable2={"name": "y", "value": 4.0}
#Variable3={"name": "e1", "value": 0.0}
#Variable4={"name": "e2", "value": 0.0}
#Variable5={"name": "e3", "value": 0.0}

#les variables de bases sont une liste de variables , et ceci est leur déclaration
variablesdebases=list[Variable]

#les variables hors bases sont une liste de variables , et ceci est leur déclaration
variableshorsbases=list[Variable]

#initialisation de variables de bases (test)
#variablesdebases= [Variable1 , Variable2]

#initialisation de variables hors bases 
#variableshorsbases= [Variable3 , Variable4, Variable5]

#La matrice provenant de l'itération est de cette forme : n lignes , m colones 
import numpy as np

n=4
m=5

mat= np.zeros((n,m))
AnalyseData(variablesdebases , variableshorsbases , mat)


import re

def validateUserInputSemantic(objectiveFunction, constraint):
    """
    Checks whether the input provided by the user is semantically correct or not
    Arguments:
        objectiveFunction : String
        constraint : String
    Returns:
        Boolean
    """
    constraint = constraint.lower()
    objectiveFunction = objectiveFunction.lower().split("=")[-1]
    constraintVariables = ['x' in constraint, 'y' in constraint, 'z' in constraint]
    objectiveFunctionVariables = ['x' in objectiveFunction, 'y' in objectiveFunction, 'z' in objectiveFunction]
    for i in range(len(constraintVariables)):
        if constraintVariables[i] and not objectiveFunctionVariables[i]: # If a variable is present in the constraint but not in the objective function
            return False
    return True

def formatUserInput(userInput):
    """
    The function takes an array of strings representing the input taken and **validated** from a jupyter user.
    PS: The function take care of a maximum of 3 variables(x, y, z) in a system i.e. R3.
    Arguments:
        userInput : List[str]
    Returns:
        Matrix : List[List[int]]
    """
    def formatVars(s):
        s = re.sub(' +', '', s.lower()) # remove white spacess
        s = re.sub('x', 'x ', s)
        s = re.sub('y', 'y ', s)
        s = re.sub('z', 'z ', s)
        s = re.sub('\+', '', s)
        return s
    
    def formatConstraints(s):
        s = re.sub('>', '', s)
        s = re.sub('<', '', s)
        s = re.sub('=', '', s)
        return s
    
    def castNumbers(inp):
        """
        The function takes a string representing a constraint or objective function.
        Arguments:
            inp : str
        Returns:
            res : List[int]
        """
        res = [0 for _ in range(nbVar + 2)] # variables coefficients + constraint + function
        for var in inp[:-2]:
            idx = 0
            if 'y' in var:
                idx = 1
            if 'z' in var:
                idx = 2
            var = var[:-1]
            if var == '-':
                res[idx] = -1
            elif var == '':
                res[idx] = 1
            else:
                res[idx] = int(var)     
        res[-2] = int(inp[-2]) # constraint row 
        res[-1] = int(inp[-1]) # function row
        return res

    def countNumberOfVariables():
        """
        The function counts the number of variables in the provided system.
        PS: Maximum number of variables is 3.
        Parameters:
            None
        Returns:
            nbVar : int
        """
        constraints = "".join(userInput[1:]).lower()
        return ('x' in constraints) + ('y' in constraints) + ('z' in constraints)
    
    # sanitize the input by removing all the extra spaces
    for i in range(len(userInput)):
        userInput[i] = re.sub(' +', ' ', userInput[i].strip())

    # get the number of variables
    nbVar = countNumberOfVariables()
    
    # create the matrix
    matrix = [[0 for _ in range(len(userInput))] for _ in range(nbVar + 2)]

    for i in range(1, len(userInput)): # skip the first line for now
        constraint = formatVars(userInput[i]).split(' ')
        flag = '0' # 0 means equal to
        if constraint[-1][0] == '>':
            # 2 means >= and 1 means >
            flag = '2' if constraint[-1][1] == '=' else '1'
        elif constraint[-1][0] == '<':
            # -2 means <= and -1 means <
            flag = '-2' if constraint[-1][1] == '=' else '-1'
        constraint.append(flag)
        constraint[-2] = formatConstraints(constraint[-2])
        res = castNumbers(constraint)
        for j in range(len(res)):
            matrix[j][i-1] = res[j]
        
    objectiveFunction = userInput[0].lower()
    op = '-1' # minimize by default
    if 'max' in objectiveFunction:
        op = '1' # 1 means maximize
    objectiveFunction = formatVars(objectiveFunction.split('=')[-1])
    objectiveFunction = [*objectiveFunction.split(' ')[:-1], '0', op]

    res = castNumbers(objectiveFunction)
    for i in range(len(res)):
        matrix[i][-1] = res[i] # add the objective function to the matrix
    return matrix

def validateUserInput(input):
    '''
    Validate user input, check if the objectif or the constraint is valid 
    '''
    #remove white spaces from the input string
    strippedInput = re.sub(' ','', input.lower())

    #get occurence of each variable
    occurenceOfX =re.sub('(max|min)[a-z]','',strippedInput).count("x")
    occurenceOfY=re.sub('(max|min)[a-z]','',strippedInput).count("y")
    occurenceOfZ =re.sub('(max|min)[a-z]','',strippedInput).count("z")

    #check if each variable exist once at most
    if(occurenceOfX > 1 or occurenceOfY > 1 or occurenceOfZ > 1):
        return False
   
    #match regEx for constraints
    matchConstraint = re.findall("^[-]{0,1}[0-9]*[x|y|z|]([+|-][0-9]*[x|y|z])*(<|>|<=|>=|=)[-]{0,1}[0-9]+$",strippedInput)
    if(len(matchConstraint)!=0):
        return True
    
     #match regEx for gain
    matchGain = re.findall("^(max|min)[a-zA-Z][=][-]{0,1}[0-9]*[x|y|z][+|-][0-9]*[x|z|y]([+|-][0-9]*[x|y|z]){0,1}$",strippedInput)
    if(len(matchGain)!=0):
        return True
   
    return False



import numpy as np
import pandas as pd
from aliases import Iterations, Variables, Variable
from utilities import Utilities
# from sympy import solve, Eq, symbols

pd.options.mode.chained_assignment = None


class Simplex():

    def __select_vars(self, init_simplex_df: pd.DataFrame) -> tuple[Variables, Variables]:
        """
        Selects the basic and non-basic variables from the initial simplex dataframe.
        """
        # TODO: what if 2 variables have the only non-zero value in the same row? which one to select?
        ret_vars = ([], [])
        for col in init_simplex_df.columns:
            if col in ['condition', 'p']:  # skip the condition and p columns
                continue
            column = init_simplex_df.loc[:, col]
            # basic variable (i.e. have only 1 non zero value)
            if column[column != 0].count() == 1:
                # get the row index of the non-zero value
                nonZeroRow = column[column != 0].index[0]
                ret_vars[0].append(
                    (col, init_simplex_df.loc[nonZeroRow]['condition'], nonZeroRow))
            else:  # non basic variable
                ret_vars[1].append((col, 0, -1))
        return ret_vars

    def __verify_solution_existence(self, variables: Variables) -> bool:
        """
        Verifies if the solution exists.
        checks if all the basic variables are positive.
        """
        for _, value, _ in variables[0]:
            if value < 0:
                return False
        return True

    def __find_equation_coef(self, iteration: pd.DataFrame) -> dict:
        equation: pd.Series = iteration.iloc[-1]
        equation = equation.drop(
            labels=["condition", "p", "ratio", "operations"])

        # equation = equation.apply(abs)
        equation = equation.astype('float64')
        min_coef_index = equation.argmin()

        return {
            "row": iteration.shape[0] - 1,
            "col": min_coef_index,
            "col_name": iteration.columns[min_coef_index]
        }

    def __find_min_ratio(self, iteration: pd.DataFrame) -> int:
        ratio = iteration["ratio"]
        ratio = ratio.drop(ratio.size - 1)  # potential pb at this line
        min_index = ratio.argmin()
        return min_index

    def __calculate_ratio(self, row: pd.Series, coef_position: dict) -> pd.Series:
        try:
            if row.name != coef_position["row"]:
                return (row["condition"] / row[coef_position["col_name"]])
            else:
                return (0)
        except ZeroDivisionError:
            return (float('inf'))

    def __transform_pivot_row(self, iteration: pd.DataFrame, pivot: dict) -> pd.DataFrame:
        iteration["operations"][pivot["position"]
                                ["row"]] = f'LP<-(1/{pivot["value"]})*LP'
        for j in range(0, iteration.shape[1] - 2):
            iteration.iloc[pivot["position"]["row"],
                           j] = iteration.iloc[pivot["position"]["row"], j] / pivot["value"]

        return iteration

    def __transform_pivot_columns(self, iteration: pd.DataFrame, pivot: dict) -> pd.DataFrame:
        for i in range(0, iteration.shape[0]):
            if i != pivot["position"]["row"] and iteration.iloc[i, pivot["position"]["col"]] != 0:
                pivot_row_coef = iteration.iloc[i, pivot["position"]["col"]]

                tmp_pivot_row = iteration.iloc[pivot["position"]["row"]]
                tmp_pivot_row = tmp_pivot_row.drop(
                    labels=["ratio", "operations"])
                tmp_pivot_row = tmp_pivot_row.apply(
                    lambda item: item * pivot_row_coef * -1)
                for j in range(0, tmp_pivot_row.size):
                    iteration.iloc[i, j] = iteration.iloc[i,
                                                          j] + tmp_pivot_row[j]

                iteration["operations"][i] = f'L{i}<-L{i}+({pivot_row_coef *-1}*LP)'

        return iteration

    def __reset_operations(self, iteration: pd.DataFrame) -> pd.DataFrame:
        iteration["operations"] = "DEFAULT"
        return iteration

    def __verify(self, iteration: pd.DataFrame) -> bool:
        done = True
        equation_row = iteration.shape[0] - 1
        cols = iteration.shape[1]
        for j in range(0, cols - 2):
            if iteration.iloc[equation_row, j] < 0:
                done = False
                break

        return done

    def __find_variable_by_equation_index(self, associated_equation: int, variables: Variables) -> Variable | None:
        target_variable: Variable = None
        for variable in variables:
            if variable[2] == associated_equation:
                target_variable = variable
                break

        return target_variable

    def __find_variable_by_name(self, name: str, variables: Variables) -> Variable:
        target_variable: Variable = None
        for variable in variables:
            if variable[0] == name:
                target_variable = variable
                break

        return target_variable

    def __update(self, iteration: pd.DataFrame, basic_variables: Variables) -> Variables:
        updated_basic_variables = []
        for variable in basic_variables:
            updated_basic_variables.append(
                (variable[0], iteration['condition'][variable[2]], variable[2]))
        return updated_basic_variables

    def _perform_simplex(self, init_simplex_df: pd.DataFrame) -> Iterations:
        variables = self.__select_vars(init_simplex_df)
        exist = self.__verify_solution_existence(variables)
        if not exist:
            raise ValueError('SIMPLEX_SOLULTION_DOES_NOT_EXIST,')

        bv = variables[0]
        nbv = variables[1]

        iteration = init_simplex_df.copy()
        ratio = [0] * iteration.shape[0]
        operations = ["DEFAULT"] * iteration.shape[0]
        iteration['ratio'] = ratio
        iteration['operations'] = operations
        iterations: Iterations = []
        iterations.append((iteration.copy(), bv, nbv))
        
        iterations_nbr = 0
        done = False
        while not done:

            # Utilities.debug(iteration, "Iteration at the start")

            iterations_nbr+= 1
            if iterations_nbr == 100 and not done:
                raise ValueError('Le programme linéaire ne converge pas.')
            # Utilities.debug(iteration, "Iteration at the start")


            # Determine the max coef position
            coef_position = self.__find_equation_coef(iteration)
            # Utilities.debug(coef_position, "The coef position")

            # Calculating ratio
            iteration["ratio"] = iteration.apply(
                self.__calculate_ratio, axis=1, args=[coef_position])
            iteration = iteration.round(5)
            # Utilities.debug(iteration, "After calculating the ratio")

            # Determine the min ratio position
            min_ratio_position = {
                "col_name": "ratio",
                "col": iteration.shape[1] - 1,
                "row": self.__find_min_ratio(iteration)
            }

            pivot_position = {
                "col": coef_position["col"],
                "row": min_ratio_position["row"]
            }

            pivot = {
                "value": iteration.iloc[pivot_position["row"], pivot_position["col"]],
                "position": pivot_position
            }
            # Utilities.debug(pivot, "calculating pivot + position")

            # Perform a transformation of the pivot row to make the pivot value equals to one
            iteration = self.__transform_pivot_row(iteration, pivot)
            iteration = iteration.round(5)
            # Utilities.debug(iteration, "Making the pivot row to 1")

            # Updating the pivot value
            pivot = {
                **pivot,
                "value": 1
            }

            # Perform transformations on the other rows to make all the values of the pivot column
            # except the pivot equal to zero
            iteration = self.__transform_pivot_columns(iteration, pivot)
            # Utilities.debug(iteration, "making the pivot to zeros")

            # Verify if we hit the final iteration of the simplex algorithm
            done = self.__verify(iteration)
            # Utilities.debug(done, "simplex stop flag")

            # Update basic and non basic variables and the iterations list
            bv = self.__update(iteration, bv)

            # Finding the variable that is going to leave the basic variables
            out_variable = self.__find_variable_by_equation_index(
                pivot["position"]["row"], bv)

            # Creation of the variable that is going to be added to the basic variables
            in_variable = (iteration.columns[pivot["position"]["col"]],
                           iteration["condition"][out_variable[2]], out_variable[2])

            bv.remove(out_variable)
            nbv.append((out_variable[0], 0, -1))

            nbv.remove(self.__find_variable_by_name(in_variable[0], nbv))
            bv.append(in_variable)

            # Utilities.debug(iteration, "Iteration at the end")
            iterations.append((iteration.copy(), bv, nbv))

            iteration = self.__reset_operations(iteration)

        return iterations


import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Slider


def drawGraph(formattedUserInput,slider=False):

    def update(val):
        idx = int(freq_slider.val)
        ax.lines.remove(objectif[0])
        a = objectifCoeff[0]
        b = objectifCoeff[1]
        c = idx
        x = np.linspace(-1000, 3000, 2)
        y=(-a*x + c)/b
        objectif[:] = ax.plot(x, y, '-g', lw=2)
        fig.canvas.draw_idle()   

    
    def drawObjectif():
        a = objectifCoeff[0]
        b = objectifCoeff[1]
        op = "Max" if objectifCoeff[3] == 1 else "Min"
        c = 0
        x = np.linspace(-1000, 3000, 2)
        y=(-a*x + c)/b
        return ax.plot(x, y, '-g',label=f"{op} Z = {a}x+{b}y",lw=2)   
    
        
    def drawConstraint(constraint):
        # ax + by = c
        a = constraint[0]
        b = constraint[1]
        c = constraint[2]
        operator = '='
        if constraint[3] == 1:
            operator = '>'
        elif constraint[3] == 2:
            operator = '>='
        elif constraint[3] == -1:
            operator = '<'
        elif constraint[3] == -2:
            operator = '<='

        colors =["#F44336","#E91E63","#9C27B0","#673AB7","#3F51B5","#2196F3","#4CAF50","#FF9800","#795548","#607D8B", ]
        randomColor = colors[np.random.randint(0, len(colors))]
        if(a == 0):# x = 0 draw horizontal line
            plt.axhline(y=c/b, color=randomColor, linestyle='-',label=f"{b}y{operator}{c}")
        elif(b == 0):# y = 0 draw vertical line
                plt.axvline(x=c/a, color=randomColor, linestyle='-',label=f"{a}x{operator}{c}")
        else:# both x and y !=0 
            x = np.linspace(0, 3000, 2)#generate 2 values for x between 1 and 10
            y=(-a*x + c)/b
            plt.plot(x, y, randomColor,label=f"{a}x+{b}y{operator}{c}")   


    def inFeasibleRegion(constraints, point):
        """
        Checks if a given point is in the feasible region.
        i.e. if the point satisfies all the constraints.
        Input:
            constraints: list of constraints
            point: tuple of the point coordinates (x, y)
        Output:
            True if the point is in the feasible region, False otherwise.
        """
        for constraint in constraints:
            a = constraint[0]
            b = constraint[1]
            c = constraint[2]
            operation = a*point[0] + b*point[1]
            if constraint[3] == 1 and operation <= c: # >
                return False
            elif constraint[3] == 2 and operation < c: # >=
                return False
            elif constraint[3] == -1 and operation >= c: # <
                return False
            elif constraint[3] == -2 and operation > c: # <=
                return False
            elif constraint[3] == 0 and operation != c: # =
                return False
        return True


    def drawIntersectionPoints(constraints):
        """
        Draws the feasible intersection points of the constraints
        """
        intersections = set()
        for i in range(len(constraints)):
            a1 = constraints[i][0]
            b1 = constraints[i][1]
            c1 = constraints[i][2]
            # draw intersection points with x axis
            if(a1 != 0):
                x = c1/a1
                intersections.add((x,0))
            # draw intersection points with y axis
            if(b1 != 0):
                y = c1/b1
                intersections.add((0,y))
                          
            for j in range(i+1,len(constraints)):
                a2 = constraints[j][0]
                b2 = constraints[j][1]
                c2 = constraints[j][2]
                try:
                    x = (b2*c1 - b1*c2)/(a1*b2 - a2*b1) 
                    y = (a1*c2 - a2*c1)/(a1*b2 - a2*b1)
                    if x >= 0 and y >= 0:
                        intersections.add((x,y))
                except ZeroDivisionError: # parallel lines
                    pass
        for intersection in intersections:
            # check if intersection is in the feasible region            
            if not inFeasibleRegion(constraints, intersection):
                continue
            plt.plot(intersection[0],intersection[1], 'o', color='black',markersize=4)
            plt.text(intersection[0],intersection[1], f"({intersection[0]:.1f}, {intersection[1]:.1f})", fontsize=8)
        
        return intersections


    def getBestAxisScaling(intersections):
        """
        Returns the best scaling for the x and y axis
        """
        maxX = 0
        maxY = 0
        for intersection in intersections:
            maxX = max(maxX, intersection[0])
            maxY = max(maxY, intersection[1])
        maxX = max(maxX, 10)
        maxY = max(maxY, 10)
        return (maxX * 1.1, maxY * 1.1)
    
    def DrawObjectifStaticly():
        a = objectifCoeff[0]
        b = objectifCoeff[1]
        x = np.linspace(0, 3000, 2)
        cRange=np.linspace(0,valmax/2,5)
        for i in range(0,len(cRange)):
            y=(-a*x + cRange[i])/b
            plt.plot(x, y, 'black',linestyle='--')  


    fig,ax = plt.subplots()
 
    #get objectif function (the objectif funtion SHOULD ALWAYS be the last int the formatted user input!)
    objectifCoeff = [row[-1] for row in formattedUserInput]
   
    listConstraints = []
    #get constraint functions
    for i in range(len(formattedUserInput[0])-1):
        listConstraints.append([row[i] for row in formattedUserInput]) 

    #draw constraint
    for constraint in listConstraints:
        drawConstraint(constraint)

    # draw intersection points
    intersections = drawIntersectionPoints(listConstraints)

    bestX, bestY = getBestAxisScaling(intersections)
    
    d = np.linspace(0, bestY, 300)
    x,y = np.meshgrid(d,d)
    x = np.linspace(0, bestX, 300)

    plt.xlim(0, bestX)
    plt.ylim(0, bestY)
    plt.xlabel(r'$x$')
    plt.ylabel(r'$y$')

    filledArea = [] 

    for i in listConstraints:
        a = i[0]
        b = i[1]
        c = i[2]
        operator = i[3]
        if(a == 0):
            if(operator == 0):
                filledArea.append(b*y==c)
            elif(operator == 1):
                filledArea.append(b*y>c)
            elif(operator == -1):
                filledArea.append(b*y<c)
            elif(operator == 2):
                filledArea.append(b*y>=c)
            elif(operator == -2):
                filledArea.append(b*y<=c)
        if(b == 0):
            if(operator == 0):
                filledArea.append(a*x==c)
            elif(operator == 1):
                filledArea.append(a*x>c)
            elif(operator == -1):
                filledArea.append(a*x<c)
            elif(operator == 2):
                filledArea.append(a*x>=c)
            elif(operator == -2):
               filledArea.append(a*x<=c)
        else:
            if(operator == 0):
                filledArea.append(b*y==c-a*x)
            elif(operator == 1):
                filledArea.append(b*y>c-a*x)
            elif(operator == -1):
                filledArea.append(b*y<c-a*x)
            elif(operator == 2):
                filledArea.append(b*y>=c-a*x)
            elif(operator == -2):
                filledArea.append(b*y<=c-a*x)

    calculate = filledArea[0]
    
    #calculate feasable area
    for i in range(1, len(filledArea)):
        calculate = calculate & filledArea[i]

    plt.imshow((calculate).astype(int), extent=(x.min(),x.max(),y.min(),y.max()), origin="lower", cmap="Greys", alpha=0.3)
    
    #draw objectif function (move the graph to see it at (0,0))
    objectif = drawObjectif()
    
    plt.grid()
    plt.legend()
    
    if(slider):
        #setup slider
        fig.subplots_adjust(left=0.25, bottom=0.25)
        axfreq = fig.add_axes([0.25, 0.1, 0.65, 0.03])
        
        # choose the best valmax for the slider
        valmax = objectifCoeff[0] * bestX + objectifCoeff[1] * bestY
        freq_slider = Slider(ax=axfreq, label='move objectif function', valmin=0, valmax=valmax, valinit=0)
        
        #listener for changes
        freq_slider.on_changed(update) 
    else:
        #draw objectif line staticly
        valmax = objectifCoeff[0] * bestX + objectifCoeff[1] * bestY
        DrawObjectifStaticly() 
    
    #open graph window
    plt.show()